/*   E-scape css inheritance and cascading
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include "libit/alloca.h"
#include <sys/types.h>
#include "libit/string.h"
#include <math.h>
#include "libtext.h"
#include "libit/malloc.h"

/* Apparently foo->(bar) in macros loses, and you have to write foo->bar
   That's contrary to common wisdom with macro parameters, but I guess
   it works OK. */

/* Return a null-terminated vector of null-terminated strings, given
   the null-terminated string `class'. */
static char **
parse_class (char *class)
{
  char **ret, *s, *p;

  ret = malloc (sizeof (char *));
  ret [0] = 0;

  while (*class) {
    while ((*class == ' ') || (*class == ','))
      class++;
    s = class;
    while ((*s != ' ') && (*s != ',') && (*s))
      s++;
    p = malloc (s - class + 1);
    memcpy (p, class, s - class);
    p [s - class] = 0;
    append_vector (&ret, p);
    class = s;
  }
  return ret;
}

/* Return the attribute in `tag' named `name' */
inline static char *
get_attr (struct text_chunk *tag, char *name)
{
  struct tag_attribute *attr;

  attr = tag->attr;

  while (attr) {
    /* FIXME should this be case sensitive? */
    if (!strcmp (attr->name, name))
      return attr->data;
    attr = attr->next;
  }
  return 0;
}

/* Return 1 if `s' is present in `superset'; else return 0. */
inline static int
present (char **superset, char *s)
{
  while (*superset) {
    /* FIXME should this be case-sensitive? */
    if (!strcmp (*superset, s))
      return 1;
    superset++;
  }
  return 0;
}

/* Return 1 if `s' appears in `possible_pseudo_name', else return 0.
   Initialize the element of `possible_pseudo_needed' at the same place
   as `possible_pseudo_name' that matched `s' to 1.  `possible_pseudo_count'
   contains the number of members in `possible_pseudo_name' and
   `possible_pseudo_needed'. */
inline static int
check_possible_pseudo (char *s, char **possible_pseudo_name,
		       int *possible_pseudo_needed,
		       size_t possible_pseudo_count)
{
  while (possible_pseudo_count) {
    /* FIXME should this be case-sensitive? */
    if (!strcmp (*possible_pseudo_name, s)) {
      *possible_pseudo_needed = 1;
      return 1;
    }
    possible_pseudo_name++;
    possible_pseudo_needed++;
    possible_pseudo_count--;
  }
  return 0;
}


/* Return 1 if `superset' contains every string in `subset', else
   return 0.  `subset' and `superset' are both null-terminated vectors.
   `possible_pseudo_name' contains a vector of names for pseudo-classes
   which might be used in the current tag.  `possible_pseudo_used'
   is a vector of ints which will be set to 1 if a different value
   would be returned by this function if the corresponding element in
   `possible_pseudo_name' were provided in `psuedo'.  `possible_pseudo_count'
   contains the number of pseudo-classes specified by `possible_pseudo_name'.
   */
static int
match_vector (char **superset, char **subset, char **possible_pseudo_name,
	      int *possible_pseudo_used, size_t possible_pseudo_count)
{
  int i;

  /* Whether everything matches so far. */
  int win = 1;

  /* An array which keeps track of which values we would set in
     `possible_pseudo_used'. */
  int *would_be_needed;

  would_be_needed = alloca (possible_pseudo_count * sizeof (int));
  for (i = 0; i < possible_pseudo_count; i++)
    would_be_needed [i] = 0;

  /* The algorithm below returns 0 immediately if something is missing
     which is not in `possible_pseudo_name'.  If something is missing
     which is in `possible_pseudo_name', `win' is set to 0, and we
     continue to see if anything else is missing or not. */

  while (*subset) {
    if (!present (superset, *subset)) {
      if (!check_possible_pseudo (*subset, possible_pseudo_name,
				  would_be_needed, possible_pseudo_count))
	return 0;
    }
    subset++;
  }
  for (i = 0; i < possible_pseudo_count; i++)
    possible_pseudo_used [i] = would_be_needed [i];
  return win;
}


/* Convert the length to pixels.  `dest' contains the length.  `parent'
   is used as the parent for percentages and `smaller' and `larger'.
   It may be provided as 0 if no conversation of percentages is
   desired; that is used for `background-position'.  A value in `mm'
   is multiplied by `base_factor' to get pixels.  `medium_base'
   contains the size of `medium' in pixels.  `medium_factor'
   is equal to `large' / `medium'.  A value in `em' will be multiplied
   by `em_factor' to get the number of pixels; likewise for `ex_factor'. */
static void
adjust_length(struct css_length *dest, struct css_length *parent,
	      float base_factor, float medium_base, float medium_factor,
	      float em_factor, float ex_factor)
{
  switch (dest->units) {
  case LENGTH_MODE_ABSOLUTE:
    dest->units = LENGTH_MODE_PX;
    dest->value = medium_base * pow((medium_factor), (dest->value));
    break;
  case LENGTH_MODE_SMALLER:
    if (parent && (parent->units == LENGTH_MODE_PX)) {
      dest->units = LENGTH_MODE_PX;
      dest->value = parent->value / medium_factor;
    };
    break;
  case LENGTH_MODE_LARGER:
    if (parent && (parent->units == LENGTH_MODE_PX)) {
      dest->units = LENGTH_MODE_PX;
      dest->value = parent->value * medium_factor;
    }
    break;
  case LENGTH_MODE_PERCENT:
    /* Do nothing when parent is 0, because for `background-position',
       the rendering code handles percentages correctly. */
    if (parent) {
      dest->units = parent->units;
      dest->value = (dest->value / 100) * parent->value;
    }
    break;
  case LENGTH_MODE_EX:
    dest->value *= ex_factor;
    dest->units = LENGTH_MODE_PX;
    break;
  case LENGTH_MODE_EM:
    dest->value *= em_factor;
    dest->units = LENGTH_MODE_PX;
    break;
  case LENGTH_MODE_IN:
    dest->value *= base_factor * 10 * 2.54;
    dest->units = LENGTH_MODE_PX;
    break;
  case LENGTH_MODE_CM:
    dest->value *= base_factor * 10;
    dest->units = LENGTH_MODE_PX;
    break;
  case LENGTH_MODE_MM:
    dest->value *= base_factor;
    dest->units = LENGTH_MODE_PX;
    break;
  case LENGTH_MODE_PT:
    dest->value *= base_factor * 10 * 2.54 / 72;
    dest->units = LENGTH_MODE_PX;
    break;
  case LENGTH_MODE_PC:
    dest->value *= base_factor * 10 * 2.54 * 12 / 72;
    dest->units = LENGTH_MODE_PX;
    break;
  case LENGTH_MODE_UNDEFINED:
  case LENGTH_MODE_NORMAL:
  case LENGTH_MODE_AUTO:
  case LENGTH_MODE_PX:
  case LENGTH_MODE_NUMBER:
  case LENGTH_MODE_TOP:
  case LENGTH_MODE_BOTTOM:
  case LENGTH_MODE_CENTER:
  case LENGTH_MODE_LEFT:
  case LENGTH_MODE_RIGHT:
  case LENGTH_MODE_BASELINE:
  case LENGTH_MODE_SUB:
  case LENGTH_MODE_SUPER:
  case LENGTH_MODE_TEXT_TOP:
  case LENGTH_MODE_TEXT_BOTTOM:
    /*  nothing  */
  }
}

/* Given the `tag' and the parent elements it points to, calculate all
   style information needed to render the tag.

   `content_style_type' should be the pointer to the `content-style-type'
   value from the http header, or 0 if no such header was specified.

   `style_sheets' contains a list of style sheets to use to determine
   the style.

   `horizontal_factor' and `vertical_factor' specify the amount to multiply
   by to convert from millimeters to pixels for horizontal and vertical
   measurements, respectively.  `medium_base' is the number of pixels in
   a `medium'-sized font.  `medium_factor' is equal to `large' / `medium'.  

   After the font size and other related properties has been determined,
   `font_size_func' will be called, with `opaque' passed intact, to
   return the values `em_factor', `ex_factor', and `font_opaque'.

   `pseudo' is a vector giving the available pseudo-classes.
   `possible_pseudo_name' contains a vector of names for pseudo-classes
   which might be used in the current tag.  `possible_pseudo_used'
   is a vector of ints which will be set to 1 if a different value
   would be returned by this function if the corresponding element in
   `possible_pseudo_name' were provided in `psuedo'.  `possible_pseudo_count'
   contains the number of pseudo-classes specified by `possible_pseudo_name'.
   */
struct css_properties *
css_cascade_and_inherit (struct text_chunk *tag, char *content_style_type,
			 struct css_sheet *style_sheets,
			 void *opaque,
			 void (*font_size_func)(void *opaque,
						struct css_properties *properties,
						float *em_factor,
						float *ex_factor,
						void **font_opaque),
			 float horizontal_factor, float vertical_factor,
			 float medium_base, float medium_factor,
			 char **pseudo,
			 char **possible_pseudo_name,
			 int *possible_pseudo_used,
			 size_t possible_pseudo_count)
{
  float em_factor, ex_factor;
  struct css_properties *properties, *parent_properties;
  char *style_string;
  struct css_sheet *current_style_sheet;
  struct css_rule *current_rule;
  struct css_selector *current_alt;
  struct css_selector *current_sel;
  struct text_chunk *current_tag;
  char **class = 0;
  int best_priority, current_priority;

  /* The first thing we do is create a priority variable for each property
     in the style sheet.  Deciding whether every rule matches gets
     extremely expensive, so we do this once, and then try matching
     properties if the selector matches. */
  /* Warning: adding a semicolon to the end of the following line will
     thoroughly break compilation.  */
#define inherit(foo) int foo ## _priority
#define inherit_border(foo) int foo ## _color_priority, \
  foo ## _width_priority, foo ## _style_priority
#include "csstypeless.h"
#undef inherit
#undef inherit_border

  parent_properties = tag->parent>properties;

  /* Allocate and initialize style */
  properties = new_css_properties ();

  /* Parse inline style if it exists. */
  /* FIXME should this be case-sensitive? */
  if ((!content_style_type) || (!strcmp (content_style_type, "text/css"))) {
    style_string = get_attr (tag, "style");
#if 0
    if (style_string)
      /* FIXME reenable this and write the function! */
      parse_inline_style (&properties, style_string);
#endif
  }

/* cssprop.in and friends rely on `inherit_int', etc.  Since we need
   one convention for determining priorities for `border', and another
   for all other properties, we use `internal_inherit_int' etc
   as the definition in this file.

   This section must come after the use of csstypeless.h, and before
   the use of any other headers.  It could be generalized if it needed
   to be, but it doesn't need to be generalized.  */

#define inherit_int(foo) internal_inherit_int(foo, foo ## _priority)
#define inherit_color(foo) internal_inherit_color(foo, foo ## _priority)
#define inherit_length(foo) internal_inherit_length(foo, foo ## _priority)
#define inherit_string(foo) internal_inherit_string(foo, foo ## _priority)
#define inherit_vector(foo) internal_inherit_vector(foo, foo ## _priority)
#define inherit_border(foo) \
  internal_inherit_length(foo.width, foo ## _width_priority); \
  internal_inherit_color(foo.color, foo ## _color_priority); \
  internal_inherit_int(foo.style, foo ## _style_priority);


  /* Initialize _priority variables properly. */

#define internal_inherit_int(foo, bar) \
  if (properties->foo) bar = PRIORITY_INLINE; else bar = 0;

#define internal_inherit_color(foo, bar) \
  if ((properties->foo.red != -1) || (properties->foo.name)) \
    bar = PRIORITY_INLINE; else bar = 0;

#define internal_inherit_length(foo, bar) \
  if (properties->foo.units) bar = PRIORITY_INLINE; else bar = 0;

#define internal_inherit_string(foo, bar) \
  if (properties->foo) bar = PRIORITY_INLINE; else bar = 0;

#define internal_inherit_vector(foo, bar) \
  if (properties->foo) bar = PRIORITY_INLINE; else bar = 0;

#include "cssprop.h"

#undef internal_inherit_int
#undef internal_inherit_color
#undef internal_inherit_length
#undef internal_inherit_string
#undef internal_inherit_vector

  for (current_style_sheet = style_sheets;
       current_style_sheet;
       current_style_sheet = current_style_sheet->next) {
    for (current_rule = current_style_sheet->rules;
	 current_rule;
	 current_rule = current_rule->next) {
      best_priority = 0;
      for (current_alt = current_rule->sel;
	   current_alt;
	   current_alt = current_alt->alt) {
	current_priority = current_sel->priority;
	if (current_priority > best_priority) {
	  current_sel = current_alt;
	  current_tag = tag;
	  while (current_tag && current_sel) {
	    if (strcasecmp (current_sel->tag, current_tag->tag_name))
	      goto lose;
	    if (current_sel->id) {
	      if (!get_attr (current_tag, "id"))
		goto lose;
	      if (strcmp (current_sel->id, get_attr (current_tag, "id")))
		goto lose;
	    }
	    class = parse_class (get_attr (current_tag, "class"));
	    if (!match_vector (class, current_sel->class, 0, 0, 0))
	      goto lose;
	    /* The pseudo-class check must be last because
	       `possible_pseudo_name' needs to be updated properly. */
	    if (!match_vector (pseudo, current_sel->pseudo,
			       possible_pseudo_name, possible_pseudo_used,
			       possible_pseudo_count))
	      goto lose;
	    current_sel = current_sel->parent;
	  lose:
	    current_tag = current_tag->parent;
	    if (class) {
	      free_vector (class);
	      class = 0;
	    }
	  }
	  if (!current_sel)
	    best_priority = current_priority;
	}
      }
      if (best_priority) {

#define internal_inherit_int(foo, bar) \
	if ((bar < best_priority) && (current_rule->properties->foo)) \
	  properties->foo = current_rule->properties->foo;

#define internal_inherit_string(foo, bar) \
	if ((bar < best_priority) && (current_rule->properties->foo)) { \
	  if (properties->foo) \
	    free (properties->foo); \
	  properties->foo = strdup(current_rule->properties->foo); \
	}

#define internal_inherit_color(foo, bar) \
	if ((bar < best_priority) \
	    && ((current_rule->properties->foo.red != -1) \
		|| (current_rule->properties->foo.name))) { \
	  if (properties->foo.name) \
	    free (properties->foo.name); \
	  copy_css_color (&properties->foo, &current_rule->properties->foo); \
	}

#define internal_inherit_length(foo, bar) \
	if ((bar < best_priority) \
	    && (current_rule->properties->foo.units)) \
	  copy_css_length (&properties->foo, &current_rule->properties->foo);

#define internal_inherit_vector(foo, bar) \
	if ((bar < best_priority) \
	    && (current_rule->properties->foo)) { \
	  if (properties->foo) \
	    free_vector (properties->foo); \
	  properties->foo = copy_vector (current_rule->properties->foo); \
	}
		
#include "cssprop.h"

#undef internal_inherit_int
#undef internal_inherit_color
#undef internal_inherit_length
#undef internal_inherit_string
#undef internal_inherit_vector

      }
    }
  }

  /* Inherit values from parent where appropriate. */

#define internal_inherit_int(foo, bar) \
  if (!bar) \
    properties->foo = parent_properties->foo;

#define internal_inherit_string(foo, bar) \
  if (!bar) \
    properties->foo = strdup (parent_properties->foo);

#define internal_inherit_color(foo, bar) \
  if (!bar) \
    copy_css_color (&properties->foo, &parent_properties->foo);

#define internal_inherit_length(foo, bar) \
  if (!bar) \
    copy_css_length (&properties->foo, &parent_properties->foo);

#define internal_inherit_vector(foo, bar) \
  if (!bar) \
    properties->foo = copy_vector (parent_properties->foo);

#include "cssinherit.h"

  /* This is a special case because of the bizare <number> handling. */
  if (!line_height_priority)
    copy_css_length (&properties->line_height,
		     &properties->line_height_inherit);

#undef internal_inherit_int
#undef internal_inherit_color
#undef internal_inherit_length
#undef internal_inherit_string
#undef internal_inherit_vector

  /* Here we resolve lengths to pixels, convert things which are relative
     to their parent to absolute units, etc. */
  /* Please be aware that the order matters.  The first length which
     must be calculated is `font_size', since all other lengths
     are calcualted relative to it.  */

  /* Arrange for the Right Thing to happen for inheriting a length.
     `foo' is the property, and `bar' is the property to use
     for a percentage.  */
#define generic_inherit_length(foo, bar, baz) \
	adjust_length(&properties->foo, bar, \
		baz, medium_base, medium_factor, em_factor, ex_factor);

  /* FIXME: this is completely bogus, but will produce the right result
     in pratice until boldness handling is really fixed... */
  if (properties->font_weight == FONT_WEIGHT_BOLDER)
    properties->font_weight = 900;
  if (properties->font_weight == FONT_WEIGHT_LIGHTER)
    properties->font_weight = 100;

  adjust_length(&properties->font_size, &parent_properties->font_size,
		 vertical_factor, medium_base, medium_factor,
		 parent_properties->em_factor, parent_properties->ex_factor);

  font_size_func(opaque, properties, &em_factor, &ex_factor,
		 &properties->font_opaque);

  properties->em_factor = em_factor;
  properties->ex_factor = ex_factor;

  if (properties->line_height.units == LENGTH_MODE_NUMBER) {
    copy_css_length (&properties->line_height_inherit,
		     &properties->line_height);
    properties->line_height.units = properties->font_size.units;
    properties->line_height.value *= properties->font_size.value;
  } else {
    generic_inherit_length (line_height, &properties->font_size,
			    vertical_factor);
    copy_css_length (&properties->line_height_inherit,
		     &properties->line_height);
  }

  generic_inherit_length (width, &parent_properties->width, horizontal_factor);
  generic_inherit_length (height, 0, vertical_factor);
  generic_inherit_length (background_position_x, 0, horizontal_factor);
  generic_inherit_length (background_position_y, 0, vertical_factor);
  generic_inherit_length (word_spacing, 0, horizontal_factor);
  generic_inherit_length (letter_spacing, 0, horizontal_factor);
  generic_inherit_length (vertical_align, &properties->line_height,
			  vertical_factor);
  generic_inherit_length (text_indent, &parent_properties->width,
			  horizontal_factor);
  generic_inherit_length (margin_top, &properties->width, vertical_factor);
  generic_inherit_length (margin_left, &properties->width, horizontal_factor);
  generic_inherit_length (margin_bottom, &properties->width, vertical_factor);
  generic_inherit_length (margin_right, &properties->width, horizontal_factor);
  generic_inherit_length (padding_top, &properties->width, vertical_factor);
  generic_inherit_length (padding_left, &properties->width, horizontal_factor);
  generic_inherit_length (padding_bottom, &properties->width, vertical_factor);
  generic_inherit_length (padding_right, &properties->width,
			  horizontal_factor);
  generic_inherit_length (border_top.width, &properties->width,
			  horizontal_factor);
  generic_inherit_length (border_left.width, &properties->width,
			  vertical_factor);
  generic_inherit_length (border_bottom.width, &properties->width,
			  horizontal_factor);
  generic_inherit_length (border_right.width, &properties->width,
			  vertical_factor);

  /* FIXME then look for relavent attributes. */

  return properties;
}

