/* Joel N. Weber II
   public domain
   This file contains generic material which needs to be included
   in any finite state machine generated by fsmbuild.  Some functions
   aren't required for all parsers, but they're short enough
   to include in all parsers, and I strongly prefer to make them
   inline.  */

/* The first section of this file includes the appropriate generic headers.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>

#if STDC_HEADERS
#include <stdlib.h>
#endif

#if WITH_DMALLOC
#include "dmalloc.h"
#endif


/* The fsm_stack_element contains a stack of states; this is used so that
   you can return to a state. */
struct fsm_stack_element {
  /* The number of the state. */
  int state;

  /* The next stack element, to be popped after the current element,
     assuming nothing is pushed in the meantime. */
  struct fsm_stack_element *next;
};


/* The struct fsm_environment stores all of the information needed by the
   a given parser. */
struct fsm_environment {
  /* The number of the state to use next time through the loop. */
  int state;

  /* A stack of states. */
  struct fsm_stack_element *stack;

  /* A string which is accumulated during parsing. */
  char *string;

  /* The size of `string' as passed to malloc. */
  size_t string_size;

  /* Number of actual valid bytes in `string'. */
  size_t string_count;
};


/* Preform initialization for `environment': initialize numbers to sane
   values and allocate space for `environment->string'. */
inline static void
fsm_clear_environment (struct fsm_environment *environment)
{
  environment->state = 0;
  environment->stack = 0;
  environment->string_count = 0;

  /* `environment->string_size' is initialized to a default which should
     be large enough for most cases.  It is grown dynamically if the
     string outgrows the initial size set here. */
  environment->string_size = 256;
  environment->string = xmalloc (environment->string_size);
}


/* Allocate memory for a new fsm_environment with `xmalloc', then initialize
   it.  The return value is the new environment.  */
inline static struct fsm_environment *
fsm_new_environment (void)
{
  struct fsm_environment *m;

  m = xmalloc (sizeof (struct fsm_environment));
  fsm_clear_environment (m);
  return m;
}


/* Free all data structures referenced by `environment', but don't free
   `environment' itself. */
inline static void
fsm_cleanup_environment (struct fsm_environment *environment)
{
  struct fsm_stack_element *stack;

  free (environment->string);
  while (environment->stack) {
    stack = environment->stack;
    environment->stack = environment->stack->next;
    free(stack);
  }
}


/* free `environment' and all the data structures `environment' refers to. */
inline static void
fsm_free_environment (struct fsm_environment *environment)
{
  fsm_cleanup_environment (environment);
  free (environment);
}


/* Push `state' on to the stack stored in `environment'. */
inline static void
fsm_push_state (struct fsm_environment *environment, int state)
{
  struct fsm_stack_element *tmp;

  tmp = xmalloc (sizeof (struct fsm_stack_element));
  tmp->next = environment->stack;
  tmp->state = state;
  environment->stack = tmp;
}


/* Pop a state off the stack contained in `environment', returning that state
   as the return value of `fsm_pop_state'. */
inline static int
fsm_pop_state (struct fsm_environment *environment)
{
  struct fsm_stack_element *tmp;
  int state;

  tmp = environment->stack;
  environment->stack = tmp->next;
  state = tmp->state;
  free (tmp);
  return state;
}


/* clear the string stored in the `environment' */
inline static void fsm_clear_string(struct fsm_environment *environment)
{
  environment->string_count = 0;
}

/* append `c' to the string stored in `environment' */
inline static void
fsm_append_char (struct fsm_environment *environment, char c)
{
  environment->string[environment->string_count] = c;
  environment->string_count++;

  if (environment->string_count == environment->string_size) {
    environment->string_size *= 2;
    environment->string = xrealloc (environment->string,
				   environment->string_size);
  }
}

/* Append the string stored in `environment' to the null-terminated string
   `s'. */
inline static void
fsm_append_to_null_string (struct fsm_environment *environment, char **s)
{
  size_t count;

  if (*s) {
    count = strlen(*s);
    *s = realloc(*s, environment->string_count + count + 1);
    memcpy((*s) + count, environment->string, environment->string_count);
    (*s)[environment->string_count + count] = 0;
  } else {
    *s = xmalloc(environment->string_count + 1);
    memcpy(*s, environment->string, environment->string_count);
    (*s)[environment->string_count] = 0;
  }
}

/* append the string stored in `environment' to the string `s' of size
   `count'  */
inline static void
fsm_append_to_sized_string (struct fsm_environment *environment, char **s,
			    size_t *count)
{
  if ((*s) && (*count)) {
    *s = realloc(*s, *count + environment->string_count);
    memcpy((*s) + (*count), environment->string, environment->string_count);
    *count += environment->string_count;
  } else {
    *s = xmalloc(environment->string_count);
    memcpy(*s, environment->string, environment->string_count);
    *count = environment->string_count;
  }
}


/* This function is a cludge to allow the string stored in environment
   to be conviniently made available as a null-terminated string.
   After calling any function which modifies the string, the string
   returned by this function will become invalid.
   This function null-terminates and returns a pointer to the string
   stored in `environment'. */
inline static char *
fsm_tmp_null_string(struct fsm_environment *environment)
{
  environment->string[environment->string_count] = 0;
  return environment->string;
}


/* Do a case-insensitive comparision between the string in
   `environment' and the null-terminated string pointed to by `s'.
   returns the same values as `strncasecmp': 0 if they are the same,
   and non-zero if they differ. */
/* FIXME: this loses if you don't have strncasecmp, or if you
   have null bytes in the middle of the string.  */
inline static int
fsm_case_cmp(struct fsm_environment *environment, char *s)
{
  size_t count;

  count = strlen(s);

  if (count != environment->string_count)
    return -1;

  return strncasecmp(environment->string, s, count);
}

