/*   libimage switch header
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* This file is the internal header for the image library.  It defines
 * several macros; the general idea is that someday I'll want to use
 * this code for something other than just building an XImage to put
 * on the screen.
 */

/* For now, we include the X11 header files.  Probably excessive, but
 * that's the easiest way.
 */

#if STDC_HEADERS
#include <stdlib.h>
#endif

#include <X11/Xlib.h>
#include <X11/Xutil.h>

struct image_struct {
  /* The actual bitmap, partially filled in.  Note that the widget
     or whatever calling this library just reads this field out
     and writes it to the screen--ugly, but it works...

     Once the image_struct is created, data points to 0 until something
     interesting can be displayed.  It gets initialized to something
     solid color. */
  XImage *data;

  /* The width and height of the image.  This information is filled
     in before the function call to create the image.  After that,
     it remains valid.	 */
  int x, y;

  /* Paramters needed by X11 color allocation */
  void *display;
  void *visual;
  int colormap;
  int depth;

  /* Used by the reader of a given format */
  void *private;

  /* Called by `image_addbytes' in imageswitch.c to add data to the image */
  void (*addbytes)(struct image_struct *image, char *data, size_t data_size);

  /* Stores a color cube for certain types of displays.
     Use `color_cube_offset' to get the offset into the cube.

     max_color_cube_index == 0 if not using the cube.

     These variables must be filled in by the application or
     widget calling the image reading library.  */
  unsigned long color_cube[125];
  int max_color_cube_index;

  /* Stores the palate for images which use one.  */
  unsigned long palate[256];
};

inline static int
color_cube_offset(int r, int g, int b, int oldmax, int newmax)
{
  return (((r * newmax) / oldmax) * newmax * newmax
	  + ((g * newmax) / oldmax) * newmax
	  + ((b * newmax) / oldmax));
}

inline static unsigned long
rgb_to_pixel(int r, int g, int b, int max, struct image_struct *image)
{
  XColor color;

  if (image->max_color_cube_index) {
    return image->color_cube[color_cube_offset(r, g, b, max,
					       image->max_color_cube_index)];
  } else {
    color.red = r;
    color.green = g;
    color.blue = b;
    XAllocColor(image->display, image->colormap, &color);
    return color.pixel;
  }
}

inline static void
set_palate(int offset, int r, int g, int b, int max,
	   struct image_struct *image)
{
  image->palate[offset] = rgb_to_pixel(r, g, b, max, image);
}

inline static unsigned long
get_palate(int offset, struct image_struct *image)
{
  return image->palate[offset];
}

inline static void
image_put_pixel_rgb(struct image_struct *image,	int x, int y,
		    int r, int g, int b, int max)
{
  XPutPixel(image->data, x, y, rgb_to_pixel(r, g, b, max, image));
}

inline static void
image_put_pixel_palate(struct image_struct *image, int x, int y, int offset)
{
  XPutPixel(image->data, x, y, get_palate(offset, image));
}

/* image_init is an internal routine which is called when the dimensions
   of the image have been successfully determined. */
void
image_init(struct image_struct *image);

/* image_new is used to create a new image.  After calling it, please
   initialize the visual, display, and depth fields yourself.  Also create
   the XCC. */
struct image_struct *
image_new(void *visual, void *display, int depth, int colormap);


/* image_free will call free() to get the image freed completely */
void image_free(struct image_struct *image);

/* image_addbyte is used to send a byte to the library for processing */
void
image_addbytes(struct image_struct *image, char *data, size_t data_size);

/* The app is expected to decided to redisplay the image whenever convenient;
 * a good time is probably right after reading data from the network, though
 * a background work procedure might be even better.
 */
