/*   E-scape network module header
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <sys/types.h>

/* Here are some data structures for identifying filetypes.
 * Four things can be used to identify files: MIME type, extension,
 * magic number, and strings.  A magic number always starts at the
 * begining of the file, while a string may appear anywhere near the
 * begining of a file.
 *
 * The priority fields are generally filled in according to the
 * following table:
 *
 * 4096 4+ byte magic numbers
 *  256 MIME types, file extensions
 *   64 3- byte magic numbers
 *   16 weak magic in html files
 *    4 text/x-gopher
 *    1 fallback of text
 *
 * Once the type has been identified, one of three things can happen:
 * either the browser knows how to handle it (ie text/html), this library
 * knows how to filter it (ie gzip), or the user supplies an external
 * program to do the filtering, which is treated as a special case
 * of #2.
 */

struct E_scape_filter {
	/*	s is the string from identification code.  For example, it
		might contain the shell command for the child filter.
		For the application callback, it contains "image",
		"text/html", or "text/plain".  (But no, it isn't
		really directly a MIME type.)  */
	char *s;

	/*	used by initial identification code and gzipfilter.c.
		data points to the first byte, and data_size holds the count */
	char *data;
	size_t data_size;

	/*	z_stream for gzip.fsm */
	void *z;

	/*	gzip.fsm uses this for fields copied directly from file */
	unsigned char method, flags;
	size_t len;

	/*	fsm state; used by at least gzip.fsm */
	void *m;

	/*	Used by the child process filter code */
	int stdin_fd, stdout_fd, stderr_fd;
	XtInputId stdout_id, stderr_id;

	/*	points back to the net structure that points to this filter */
	struct net_connection *net;

	/*	function to call with data processed by this filter.
		to cleanup, call filter->app_data(filter->next, 0, 0);  */
	void (*app_data)(struct E_scape_filter *filter, char *buf,
		size_t count);

	/*	next filter in chain */
	struct E_scape_filter *next;
};

struct identify_info {

	/*	sample data to match against */
	char *data;

	/*	test for data & mask equal to something & mask */
	char *mask;

	/*	length of data and mask */
	size_t length;

	/*	priority of this matching info.  FIXME how does this work? */
	int priority;

	struct identify_info *next;
};

struct type_info {
	struct identify_info *mime, *ext, *magic, *string;
	void (*app_data)(struct E_scape_filter *filter, char *buf, size_t count);
	char *s;
	struct type_info *next;
};

void E_scape_ident_init(void);
struct type_info *E_scape_ident_find_match(char *mime, char *ext,
	char *data, int data_size);
void _E_scape_net_ident(struct E_scape_filter *filter, char *buf, size_t count);
struct E_scape_filter *_E_scape_new_filter(struct net_connection *net);



struct url {
  /* URLs are in the form
     protocol://username:password@hostname:port/filename#position
     */

  /* The entire raw URL */
  char *url;

  /* The name of the protocol, such as "html" */
  char *protocol;

  /* The entire username, or 0.  For example, "nemo". */
  char *username;

  /* The password, such as "password" */
  char *password;

  /* The hostname, such as "nautilus.sub". */
  char *hostname;

  /* port for network conection; such as 80 */
  int port;

  /* The entire filename, such as "/~nemo/foo.gz" */
  char *filename;

  /* The `client-side specifier'.  For example, in the URL
     "http://www.foo.com/bar#baz", this field contains
     "baz".  If the URL were "http://www.foo.com/bar#baz#mumble",
     then this field would store "baz#mumble", but FIXME how
     do we process it from there? */
  char *position;

  /* A literal error message to be displayed, in case the URL
     parser wants to report an error.  Null terminated. */
  char *error;

  /* An open file descriptor, used when the file exists on the localhost,
     opened by the url parser. */
  int fd;
};

struct net_connection {
	/*	The URL we're fetching. */
	struct url url;

	/*	FIXME It isn't clear what this is. */
	char *auth_url;

	/*	Contains info for POST method.  FIXME Not sure what the auth
		side is for.  */
	int post_size, auth_post_size;
	char *post_data, *auth_post_data;

	/*	Non-zero if we should avoid caching the file and tell the
		proxy not to cache the file. */
	int no_cache;

	/*	FIXME doesn't seem to get initialized */
	/*	Counts the number of times we process HTTP redirects,
		to detect infinite loops.  Initialized to ??? and then
		decrememted to 0.  */
	int redir_count;

	/*	FIXME what are these? */
	char *auth_challenge;
	char *auth_protocol;
	char *auth_realm;
	char *location;

	/*	File descriptors and select IDs for various network
		connections.  net_fd is used as the primary file descriptor;
		data_fd == net_fd initially, but for FTP they're different.
		listen_fd is probably only used for FTP.  cache_fd is
		currently not used. */
	int cache_fd, net_fd, data_fd, listen_fd;
	XtInputId net_id, data_id, listen_id;

	/* Function which is called when the connection has been established.
	 * This function is contained in the network code; there is one
	 * version for each protocol.
	 */
	void (*connection_done)(struct net_connection *net);

	/*	app_data is the output of the low level network code.
		out_data is used to pass data to the application program.
		Yes, this does seem backwords; but that's what happens
		when I don't know what I'm doing.  */
	void (*app_data)(struct E_scape_filter *filter, char *buf,
		size_t count);
	void (*out_data)(struct E_scape_filter *filter, char *buf,
		size_t count);

	/*	status_func is called with any informative messages.
		Generally, this will include everything the FTP server
		says on the control connection, as well as various
		other status messages.  The information is displayed
		in the status line at the bottom of the browser window.
	 */
	void (*status_func)(struct net_connection *net, char *mesg);

	/*	current_size is accurate.  total_size is as accurate as the
		network protocols allow it to be.  There is some logic which
		fudges to keep total_size >= current_size.  */
	int total_size, current_size;

	/* The mime type provides one method for identifying the type
	 * of data being retrieved.  This file is guarenteed to be filled
	 * in by the time app_data() is called if it will get filled in.
	 * finger sets it to text/plain, ftp leaves it null, http sets it
	 * when it knows, and localhost files get it left null.  For gopher,
	 * it will be text/x-gopher.  Note that
	 * I haven't yet figured out the exact rules for determining file
	 * types, and a priority scheme will be used in which certain mime
	 * types are extremely high priority (ie text/plain and text/html),
	 * and certain magic numbers are the next level of importance
	 * (GIF89a, for example), and extensions will only be used most
	 * likely for .html and .htm, since most file formats have
	 * good magic numbers.  Then a fuzzy search for <HTML> and <HEAD>
	 * and <BODY> and <TITLE> will be done to see if the file is
	 * likely to be html.  If these suggest html, then html is
	 * used.  Otherwise we have plain ASCII text.
	 *
	 * Also note that I haven't figured out the exact rules for handling
	 * files embeded in .tar.gz, since mime_type would in this case
	 * apply to the tar.
	 */
	/*	FIXME I'm thinking in C right now, and haven't bothered
		groking and updating the above comment.  I want to make
		this be a linked list.  You can have data: urls where
		you have to do base64 decoding and gunzip before getting
		usable data.  */
	char *mime_type;
	char *mime_encoding;

	/*	The first filter to pass recieved data to.  It in turn
		will pass it to other filters if nessisary.  */
	struct E_scape_filter *filter;

	/*	I use void because I don't feel like including the rather large
		X headers for e-scape-get.  */
	void *app_context;
	void *read_condition, *write_condition;

	/*	Since the precise details of writing data to the network and
		retrying later if the buffer is full are a bit complex,
		we use some standardized code.  See netwrite.c to
		understand this.  */
	char *write_block_start, *write_block_mark, *write_block_end;
	int write_fd;
	XtInputId write_id;

	/*	Info about DNS child process.  */
	pid_t dns_pid;
	int dns_out, dns_err;
	XtInputId dns_out_id, dns_err_id;

	/* The app owns the following value.  It will be initialized
	 * to 0
	 */
	void *opaque;

	/* Used by http code */
	char *mime_field, *mime_info;
	int mime_state;
	char mime_nl;

	/* Used by the ftp code */
	char *ftp_buf;
	size_t ftp_buf_size, ftp_buf_count;
	void (*ftp_response) (struct net_connection *net);
	int ftp_response_code;
	char port_cmd[32];
};

void E_scape_net_stop(struct net_connection *net);
void E_scape_net_free(struct net_connection *net);
struct net_connection *E_scape_net_init(
	void (*app_data)(struct E_scape_filter *filter, char *buf,
		size_t count),
	void *app_context, void *read_condition, void *write_condition);
void E_scape_net_goto(struct net_connection *net, char *url,
	char *post_data, size_t post_size, int no_cache);
void E_scape_zombie_init(void);

void _E_scape_net_done(struct net_connection *net);
void _E_scape_net_go(struct net_connection *net);
void _E_scape_net_write(struct net_connection *net, char *data, size_t count);
void _E_scape_net_input_install(struct net_connection *net);
void _E_scape_net_html_error(struct net_connection *net, char *msg, int err);
void _E_scape_connected_finger(struct net_connection *net);
void _E_scape_connected_gopher(struct net_connection *net);
void _E_scape_connected_ftp(struct net_connection *net);
void _E_scape_connected_http(struct net_connection *net);
void _E_scape_connected_http_0_9(struct net_connection *net);
void _E_scape_net_make_auth_form(struct net_connection *net);

void _E_scape_net_child_filter(struct E_scape_filter *filter, char *buf,
	size_t count);
void _E_scape_net_gunzip_filter(struct E_scape_filter *filter, char *buf,
	size_t count);
void _E_scape_net_deflate_filter(struct E_scape_filter *filter, char *buf,
	size_t count);

void _E_scape_net_url_zero(struct url *url);
void _E_scape_net_url_cleanup(struct url *url);
void _E_scape_net_url_parse(struct url *url, char *URL);

/*	Contains the number of bytes recieved since it was last zero'd.
	The graph in the X11 client reads this value and zeros it,
	usually once per second.  */
extern int bytes_transfered;
