/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* This file contains a finite state machine parser for ppm files.  There
 * are a few conventions to be aware of: I have a habit of refering to
 * a ppm data structure when I really mean pnm, but I'm very precise when
 * it comes to function names: a ppm function deals with color, which a pgm
 * function handles grayscale.
 *
 * I would not be surprised if there are bugs.  This parser supports 6
 * different formats which have similar headers.  If you can send me
 * an image which this code has trouble with, I will look at it.
 * Should take me an hour at most to fix.
 *
 * Incidentally, the format that this file reads is defined by the
 * netpbm package.  There's a manpage in it somewhere.
 */

#include <config.h>
#include "image.h"

#define POS_WIDTH  0
#define POS_HEIGHT 1
#define POS_MAX    2
#define POS_DATA   3

#define COLOR_RED   0
#define COLOR_GREEN 1
#define COLOR_BLUE  2

#define LWS_LWS     0
#define LWS_DATA    1
#define LWS_COMMENT 2

struct ppm_struct {
	/* The maximum value of numbers for ppm and pgm
	 */
	int max;

	/* The current location in the picture
	 */
	int x;
	int y;

	/* Used when a given number has been half parsed.?
	 */
	int color;

	/* See LWS_ #defines above
	 */
	int lws;

	/* Probably used for temporary accumluation of data
	 */
	int red;
	int green;
	int blue;
	int acc;

	/* A function called by the generic code that reads an int
	 */
	void (*handle_acc) (struct image_struct *image, int c);

	/* '1'-'6' reflecting the P1, P2, P3, etc at top of file.
	 */
	char format;
};

/* Routine we use when we have given up.
 */
static void devnull_add(struct image_struct *image,
	char *data, size_t data_size)
{
}

/* The next six routines handle different types of p?m files.  There
 * are six different formats; portable PIXEL maps (color); portable
 * GRAY maps (grayscale); portable BIT maps (black + while).  Some
 * are human readable (well, almost), and some are binary.  This gives
 * six different permutations.  The only interesting thing to note
 * is that the raw functions take the data directly, while the txt
 * functions allow the generic accumulator to handle the details.
 */
static void ppm_raw_add(struct image_struct *image,
	char *data, size_t data_size)
{
	struct ppm_struct *ppm = image->private;

	while (data_size) {
		switch (ppm->color) {
			case COLOR_RED:
				ppm->red = *data;
				ppm->color = COLOR_GREEN;
				break;
			case COLOR_GREEN:
				ppm->green = *data;
				ppm->color = COLOR_BLUE;
				break;
			case COLOR_BLUE:
				image_put_pixel_rgb(image, ppm->x, ppm->y,
					ppm->red, ppm->green, *data, ppm->max);
				ppm->color = COLOR_RED;
				ppm->x++;
				if (image->x == ppm->x) {
					ppm->x = 0;
					ppm->y++;
				}
				if (ppm->x == image->x) {
					image->addbytes = devnull_add;
					return;
				}
				break;
		}
		data++, data_size--;
	}
}


static void pgm_raw_add(struct image_struct *image,
	char *data, size_t data_size)
{
	struct ppm_struct *ppm = image->private;

	while (data_size) {
		image_put_pixel_rgb(image, ppm->x, ppm->y, *data, *data, *data,
			ppm->max);
		ppm->x++;
		if (ppm->x == image->x) {
			ppm->x = 0;
			ppm->y++;
		}
		if (ppm->y == image->y) {
			image->addbytes = devnull_add;
			return;
		}
		data++, data_size--;
	}
}


static void pbm_raw_add(struct image_struct *image,
	char *data, size_t data_size)
{
	struct ppm_struct *ppm = image->private;
	int i;

	while (data_size) {
		for (i = 0; i <= 7; i++) {
			if ((*data) & 0x80)
				image_put_pixel_rgb(image, ppm->x, ppm->y,
					1, 1, 1, 1);
			else
				image_put_pixel_rgb(image, ppm->x, ppm->y,
					0, 0, 0, 1);
			*data <<= 1;
			ppm->x++;
			if (ppm->x == image->x) {
				ppm->x = 0;
				ppm->y++;
				if (ppm->y == image->y) {
					image->addbytes = devnull_add;
					return;
				}
			}
		}
		data++, data_size--;
	}
}


static void ppm_txt_add(struct image_struct *image, int c)
{
	struct ppm_struct *ppm = image->private;

	switch (ppm->color) {
		case COLOR_RED:
			ppm->red = c;
			ppm->color = COLOR_GREEN;
			break;
		case COLOR_GREEN:
			ppm->green = c;
			ppm->color = COLOR_BLUE;
			break;
		case COLOR_BLUE:
			ppm->color = COLOR_RED;
			image_put_pixel_rgb(image, ppm->x, ppm->y, ppm->red,
				ppm->green, c, ppm->max);
				ppm->x++;
				if (ppm->x == image->x) {
					ppm->x = 0;
					ppm->y++;
				}
				if (ppm->y == image->y)
					image->addbytes = devnull_add;
			break;
	}
}


static void pgm_txt_add(struct image_struct *image, int c)
{
	struct ppm_struct *ppm = image->private;

	image_put_pixel_rgb(image, ppm->x, ppm->y, ppm->red,
		ppm->red, ppm->red, ppm->max);
	ppm->x++;
	if (ppm->x == image->x) {
		ppm->x = 0;
		ppm->y++;
	if (ppm->y == image->y)
		image->addbytes = devnull_add;
	}
}


static void pbm_txt_add(struct image_struct *image, int c)
{
	struct ppm_struct *ppm = image->private;

	if ((c == '0') || (c == '1')) {
		if (c == '0')
			image_put_pixel_rgb(image, ppm->x, ppm->y, 1, 1, 1, 1);
		else
			image_put_pixel_rgb(image, ppm->x, ppm->y, 0, 0, 0, 0);
		ppm->x++;
		if (ppm->x == image->x) {
			ppm->x = 0;
			ppm->y++;
			if (ppm->y == image->y) {
				image->addbytes = devnull_add;
			}
		}
	}
}


/* These functions parse the header for all files.  The only differences
 * between the headers of the different files are that bitmaps need no
 * max value because black and white are the only possible value, and
 * the second byte of the file chooses the format.
 */
static void handle_max(struct image_struct *image, int c)
{
	struct ppm_struct *ppm = image->private;

	ppm->max = c;
	if (ppm->format == '2')
		ppm->handle_acc = pgm_txt_add;
	else if (ppm->format == '3')
		ppm->handle_acc = ppm_txt_add;
	else if (ppm->format == '5')
		image->addbytes = pgm_raw_add;
	else /* Assume format 6 */
		image->addbytes = ppm_raw_add;
}


static void handle_height(struct image_struct *image, int c)
{
	struct ppm_struct *ppm = image->private;

	image->y = c;
	ppm->handle_acc = handle_height;
	image_init(image);
	if (ppm->format == '1') {
		ppm->handle_acc = pbm_txt_add;
	} else if (ppm->format == '4') {
		image->addbytes = pbm_raw_add;
	} else {
		ppm->handle_acc = handle_max;
	}
}

static void handle_width(struct image_struct *image, int c)
{
	struct ppm_struct *ppm = image->private;

	image->x = c;
	ppm->handle_acc = handle_height;
}


/* This function handles the process of converting ASCII numbers to binary
 * numbers.  It also strips comments.  It calls other functions when it
 * finds a complete number.
 */
static void acc_add(struct image_struct *image,
	char *data, size_t data_size)
{
	struct ppm_struct *ppm = image->private;

	while (data_size) {
		if ((ppm->acc != -1) && ((*data < '0') || (*data > '9'))) {
			data++, data_size--;
			(*ppm->handle_acc)(image, ppm->acc);
			ppm->acc = -1;
			if (image->addbytes != acc_add) {
				(*(image->addbytes))(image, data, data_size);
				return;
			}
		}
		if (*data == '#')
			ppm->lws = LWS_COMMENT;
		if (ppm->lws == LWS_COMMENT) {
			while (data_size && (*data != '\n') && (*data != '\r'))
				data++, data_size--;
			if ((*data == '\n') || (*data == '\r'))
				ppm->lws = LWS_LWS;
		} else if ((*data >= '0') && (*data <= '9')) {
			if (data_size && (ppm->acc == -1))
				ppm->acc = 0;
			while (data_size && (*data >= '0') && (*data <= '9')) {
				ppm->acc *= 10;
				ppm->acc += *data - '0';
				data++, data_size--;
			}
		} else
			data++, data_size--;
	}
}


/* initialization function (sort of)
 */
void image_addbytes_ppm(struct image_struct *image,
	char *data, size_t data_size)
{
	struct ppm_struct *ppm;

	/* The following code means that you must pass at least 2 bytes in
	 * the begining; I hope this never becomes a problem.
	 */
	if ((data_size < 2) || (*data != 'P')) {
		image->addbytes = devnull_add;
		return;
	}

	ppm = malloc(sizeof(struct ppm_struct));

	/* max is initialized later.  red green blue are temp storage */

	ppm->x = 0;
	ppm->y = 0;
	ppm->color = 0;
	ppm->lws = 0;
	ppm->handle_acc = handle_width;
	ppm->acc = -1;

	image->private = ppm;
	image->addbytes = acc_add;

	data++;
	ppm->format = *data;
	data++;
	data_size -= 2;

	(*(image->addbytes))(image, data, data_size);
}


/* identify a p?m file.  Called by switch.c; even though the generic
 * identification code works, it doesn't bother explaining its results
 * to the graphics code.  may or may not change later
 */
int image_identify_ppm(char *data, size_t data_size)
{
	if (data_size < 2) return 0;
	if (*data != 'P') return 0;
	data++;
	if (*data > '6') return 0;
	if (*data < '1') return 0;
	return 1;
}
