/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include <X11/Xlib.h>
#include <math.h>
#include "libtext.h"
#include "image.h"

/* The background drawing functions work on rectangular areas.
   Unfortunately, some paragraphs are non-rectangular.  So
   `draw_main_background' draws the central area where the image
   will be rendered, and `draw_aux_background' is used to draw other
   areas.  (If repeating of the image is enabled, however, the aux
   area can also have image in it.) */

/* Given the length `l', and the `background_size' and `image_size'
   in pixels, return the number of pixels by which the upper-left
   hand corner of the image will be offset from the upper-left
   hand corner of the background, for one dimension.  Note that the
   return value can be negative. */
inline static int
find_axis_origin (struct css_length *l, int background_size, int image_size)
{
  switch (l->units) {
  case LENGTH_MODE_PX:
    return rint (l->value);
  case LENGTH_MODE_PERCENT:
    return rint ((l->value * background_size)
		 - ((100 - l->value) * image_size));
  case LENGTH_MODE_TOP:
  case LENGTH_MODE_LEFT:
    return 0;
  case LENGTH_MODE_BOTTOM:
   case LENGTH_MODE_RIGHT:
    return background_size - image_size;
  default:
    return (background_size - image_size) / 2;
  }
}

/* Given `properties', and a background whose size is specified by
   `background_width' and `background_height', with an image
   whose size is specified by `image_width' and `image_height', return
   the origin of the background image in `x' and `y'.  */
static void
find_image_origin (struct css_properties *properties,
		   int background_width, int background_height,
		   int image_width, int image_height,
		   int *x, int *y)
{
  *x = find_axis_origin (&properties->background_position_x, background_width,
			 image_width);
  *y = find_axis_origin (&properties->background_position_y, background_height,
			 image_height);
}

/* Return the location where the upper-left hand corner of the first
   image in the repeating series would be if the whole image were
   drawn on the screen.  `origin' is the upper-left hand corner of the
   main image specified by `background-position'; `image' is the width
   or height of the image; and `background' is the upper-left hand
   corner of the image. */
inline static int
repeat_starting_value (int origin, int image, int background)
{
  while (origin < background)
    origin += image;
  while (origin > background)
    origin -= image;
  return origin;
}

/* Set `*width' to the width of `image' in pixels, and set `*height'
   to the height of `image' in pixels. */
inline static void
get_image_size (struct image_struct *image, int *width, int *height)
{
  *width = image->x;
  *height = image->y;
}

/* Draw the background specified by `properties' to the rectangle bounded by
   (`x1', `y1') and (`x2', `y2'), inclusive. */
static void
draw_background_color (Display *display, Drawable d,
		       GC gc, unsigned long *gc_foreground,
		       struct css_properties *properties,
		       int x1, int y1, int x2, int y2)
{
  if (properties->background_color.pixel != *gc_foreground) {
    XSetForeground (display, gc, properties->color.pixel);
    *gc_foreground = properties->color.pixel;
  }
  XFillRectangle (display, d, gc, x1, y1, x2 + 1 - x1, y2 + 1 - y1);
}

/* Draw the background specified by `properties' to `display' and `d' using
   `gc'.  `gc_foreground' contains a pointer to the current foreground.
   `x' and `y' specify the origin of the image; the area to draw to
   is bounded by (`x1', `y1') and (`x2', `y2'), inclusive. */
/* FIXME this ignores the alpha channel. */
static void
draw_one_background_image (Display *display, Drawable d,
			   GC gc, unsigned long *gc_foreground,
			   struct css_properties *properties,
			   int x, int y, int x1, int y1, int x2, int y2)
{
  int image_width, image_height;

  get_image_size (properties->background_image, &image_width, &image_height);

  while (x > x1)
    x -= image_width;
  while ((x + image_width) < x1)
    x += image_width;
  while (y < y1)
    y -= image_height;
  while ((y + image_height) < y1)
    y += image_height;

  XPutImage (display, d, gc, properties->background_image->data, x1 - x, y1 - y,
	     x1, y1, x2 + 1 - x1, y2 + 1 - y1);
}

/* Draw the background specified by `properties' in the box including
   (`x1', `y1') as its upper-left hand corner and (`x2', `y2')
   as its lower-right hand corner.  Draw to `display' and `d'
   using `gc'.  `gc_foreground' contains the current value of the
   foreground, and will be updated if `gc's foreground changes.
   The main region is bounded by (`x3', `y3') and (`x4', `y4');
   it is used to align the image properly. */
void
draw_aux_background (Display *display, Drawable d,
		     GC gc, unsigned long *gc_foreground,
		     struct css_properties *properties,
		     int x1, int y1, int x2, int y2,
		     int x3, int y3, int x4, int y4)
{
  int origin_x, origin_y, image_width, image_height, x, y;

  if (!properties->background_image) {
    draw_background_color (display, d, gc, gc_foreground, properties,
			   x1, y1, x2, y2);
    return;
  }

  get_image_size (properties->background_image, &image_width, &image_height);
  find_image_origin (properties, x4 - x3, y4 - y3, image_width, image_height,
		     &origin_x, &origin_y);
  if (properties->background_repeat | BACKGROUND_REPEAT_X) {
    if (properties->background_repeat | BACKGROUND_REPEAT_Y) {
      for (x = repeat_starting_value (origin_x, image_width, x1);
	   (x + image_width) < x2; x += image_width) {
	for (y = repeat_starting_value (origin_y, image_height, y1);
	     (y + image_height) < y2; y += image_height) {
	  draw_one_background_image (display, d, gc, gc_foreground, properties,
				     x, y, x1, y1, x2, y2);
	}
      }
    } else {
      for (x = repeat_starting_value (origin_x, image_width, x1);
	   (x + image_width) < x2; x += image_width) {
	draw_one_background_image (display, d, gc, gc_foreground, properties,
				   x, origin_y, x1, y1, x2, y2);
      }
      draw_background_color (display, d, gc, gc_foreground, properties,
			     x1, y1, x2, origin_y - 1);
      draw_background_color (display, d, gc, gc_foreground, properties,
			     x1, origin_y + image_height + 1, x2, y2);
    }
  } else {
    if (properties->background_repeat | BACKGROUND_REPEAT_Y) {
      for (y = repeat_starting_value (origin_y, image_height, y1);
	   (y + image_width) < y2; y += image_height) {
	draw_one_background_image (display, d, gc, gc_foreground, properties,
				   origin_x, y, x1, y1, x2, y2);
      }
      draw_background_color (display, d, gc, gc_foreground, properties,
			     x1, y1, origin_x - 1, y2);
      draw_background_color (display, d, gc, gc_foreground, properties,
			     origin_x + image_width + 1, y1, x2, y2);
    } else {
      draw_one_background_image (display, d, gc, gc_foreground, properties,
				 origin_x, origin_y, x1, y1, x2, y2);
      draw_background_color (display, d, gc, gc_foreground, properties,
			     x1, y1, origin_x - 1, y2);
      draw_background_color (display, d, gc, gc_foreground, properties,
			     origin_x + image_width + 1, y1, x2, y2);
      draw_background_color (display, d, gc, gc_foreground, properties,
			     origin_x, y1, origin_x + image_width,
			     image_height - 1);
      draw_background_color (display, d, gc, gc_foreground, properties,
			     origin_x, image_height + 1,
			     origin_x + image_width, y2);
    }
  }
}

/* Draw the background specified by `properties' in the box including
   (`x1', `y1') as its upper-left hand corner and (`x2', `y2')
   as its lower-right hand corner.  Draw to `display' and `d'
   using `gc'.  `gc_foreground' contains the current value of the
   foreground, and will be updated if `gc's foreground changes. */
void
draw_main_background (Display *display, Drawable d,
		      GC gc, unsigned long *gc_foreground,
		      struct css_properties *properties,
		      int x1, int y1, int x2, int y2)
{
  draw_aux_background (display, d, gc, gc_foreground, properties,
		       x1, y1, x2, y2, x1, y1, x2, y2);
}
